/*===============================<o>=====================================

  This software was written by Ian L. Kaplan, Chief Fat Bear, Bear Products
  International.  Use of this software, for any purpose, is granted on two
  conditions: 

   1. This copyright notice must be included with the software
      or any software derived from it.

   2. The risk of using this software is accepted by the user.  No
      warranty as to its usefulness or functionality is provided
      or implied.  The author and Bear Products International provides
      no support.

 ===============================<o>=====================================*/

#include <stdio.h>
#include "freelist.h"

/* This class supports an unordered list of elements, where all elements
   are assumed to be the same.  Specifically, it supports a "free list"
   for a memory allocater, where all the elements in the free list
   are the same size.

   When items are removed from the list, this object keeps the list
   "links" for reuse.  This reduces the number of times that this
   object must request more memory when items are frequently added
   and removed from the list.  But it can be a problem if many items
   are added to the list and then removed, since none of the list
   memory will be returned to the memory allocater for reuse.

*/


/* add_to_free_list

   This function adds "addr" to the free list.  When an item is added
   to a list, it is necessary to allocate a "link" in the list to
   store information about the item.  If there is a free "link"
   already available from an item that has been previously removed
   from the list, one of these will be used for the new item.
   Otherwise a new "link" will be allocated.  Once a new linke is
   available, initialize the link with "addr" and add it to the list.

   chain_head : a list of free "links"
   list_head  : the list that the new item will be added to

*/
void freelist::add_to_free_list( void *addr )
{
  list *new_link;

  if (addr != NULL) { // don't try to put null pointers in the free list
    if (chain_head == NULL) {
      new_link = (list *)get_mem( sizeof( list ));
    }
    else {
      new_link = chain_head;
      chain_head = chain_head->next;
    }
    
    if (new_link != NULL) {
      // Initialize the "link" and add it to the free list
      new_link->addr = addr;
      new_link->next = list_head;
      list_head = new_link;
    }
  }
}  // freelist::add_to_free_list



/* get_from_free_list

   Get an item from the free list or return NULL if the list is empty.
   If the item is retrieved from the free list, add the now unused "link"
   to the "chain" list, so that it can be reused later, when another
   item is added to the list.

 */
void *freelist::get_from_free_list(void)
{
  void *new_elem = NULL;

  if (list_head != NULL) {
    list *free_link = list_head;

    // get the address of the free memory and unlink the list "link".
    new_elem = list_head->addr;
    list_head = list_head->next;

    // Add the list link to the free link chain
    free_link->next = chain_head;
    chain_head = free_link;
  }
  
  return new_elem;
} // freelist::get_from_free_list


//
// get_mem
//
// Memory allocater for the freelist object.  The derived class must
// supply this function.
//
void *freelist::get_mem( unsigned int num_bytes )
{
  printf("freelist::get_mem: no memory allocation function for class\n");
  return (void *)NULL;
}
