
#ifndef HASH_H

#define HASH_H

#include "freelist.h"
#include "bitree.h"

/*===============================<o>=====================================

  This software was written by Ian L. Kaplan, Chief Fat Bear, Bear Products
  International.  Use of this software, for any purpose, is granted on two
  conditions: 

   1. This copyright notice must be included with the software
      or any software derived from it.

   2. The risk of using this software is accepted by the user.  No
      warranty as to its usefulness or functionality is provided
      or implied.  The author and Bear Products International provides
      no support.

 ===============================<o>=====================================*/

/* 
   This object implements a hash table using a binary tree.  The
   advantage of this implementation is that it is more memory
   efficient than a hash algorithm that uses an array.  The tree has
   no wasted space, as the array does.  The disadvantage is
   performance.  Hash accesses are slower than they would be for a
   balanced tree and considerably slower than an array based hash
   table (although this object is still faster than the Microsoft VC++
   malloc memory allocator).

   If the binary tree that this hash function is based on were always
   perfectly balanced, one access to the tree would require log2(N)
   operations, on average (where N is the number of elements in the
   tree).  For M access, this will be M * log2(N), which is better
   than N^2, but not nearly as good as M for a true hash table.  In
   practice, the performance is even worse.  Balanced binary trees
   (AVL trees) require a fair amount of code to implement, so there is
   a computation time cost.  If the tree is not an AVL tree (which is
   always balanced), the keys can be randomized (see the hash_value
   function).  However, for sequentially increasing keys, it is very
   difficult to come up with a function that will map a set of
   sequential numbers into a set of pseudo-random numbers.  So for a
   set of sequential numbers the tree will have branches much longer
   than log2(N) (the max_branch_size function in the binary_tree
   object can be used to print the maximum branch for a given
   tree). 

   This class uses the freelist object to keep track of deallocated
   tree nodes.  This class is inherited, rather than used as a class
   variable, because its memory allocater (get_mem) is a virtual
   function.  Using a virtual function for the memory allocater allows
   the user of the class to decide how memory will be allocated.

*/


class hash : public binary_tree, public freelist
{
private: // class functions
  unsigned int hash_value( unsigned int val );

public: // class functions
  // hash(void) { }
  hash(void) : freelist( sizeof( bi_tree) ) {}
  void put_in_hash(void *addr, unsigned int num_bytes )
    { tree_insert(hash_value((unsigned int)addr), num_bytes); }

  unsigned int get_from_hash( void *addr )
    { return tree_delete(hash_value((unsigned int)addr) ); }

  binary_tree::bi_tree_ptr get_tree_mem(void);

  void free_tree_mem(binary_tree::bi_tree_ptr node)
    { add_to_free_list( (void *)node ); }

  virtual void *get_mem( unsigned int num_bytes );
}; // class hash


#endif
